#!/usr/bin/perl -w
use diagnostics ;
use strict ;

use ezApp::DB ;	

package ccKernel ;		# Handles transactions and queries on community currency databases.
					
					#	Copyright (c) 2001 Richard K. Moore.  All Rights Reserved.
					
#-----------------------------------------------------------------------------------------------------------
#
#		Except for &new(), failures always return '', and successful results test as true.
#		Diagnostic messages are sent to $job->Warn().
#
#		The modules CGIS::DB and CGIS::Localizer are required by this module.  CGIS::DB uses the DBI interface 
#		and expects to be using mySQL.  CGIS::Localizer contains the code which opens the database, and the code 
#		which sends email messages.  These operations are typically installation specific, and Localizer.pm should 
#		be customized as necessary.
#		
#		ccKernel.pm is based on a mySQL database which includes at least these base tables:
#
#			( Currencies, Origins, Ledgers, Acks )
#		
#		The tables must contain these columns, in any order:
#	
#			Currencies: 	CurrencyId
#			Origins:			AcctId, AcctPassword
#			Ledgers:			AcctId, CurrencyId, Balance, TradesTotal, TradesCount, LastTradeDate
#			Acks:					AckDateTime, SourceAcctId, DestAcctId, AckCurrencyId, AckAmount, AckItem, SourceBalance, DestBalance
#
#		The file 'CreateMinCC.sql' provides mySQL statements which will create a valid, minimum database ('ccMin').
#
#		All amounts are stored as integers.  Client scripts can support a decimal interace to users if that is desired,
#		and interface to ccKernel in terms of the minimal units (cents, pence, or whatever).
#
#
#-----------------------------------------------------------------------------------------------------------
# CHANGE LOG
#
# 4/6/02 - rkm
#		- &new: Accept as args: $DBHandler, $DBName, $DBUser, $DBPassword
#
#-----------------------------------------------------------------------------------------------------------
sub new 						# Instantiate and  return self.
										# Establish CurrencyOwnerIds and UserPasswords.
										# Set $self->{isopen} to false if problems, else true.
{ 
	my $class = shift;
	my ( $job, $DBHandler, $DBName, $DBUser, $DBPassword ) = @_;
#	die "ccKernel new: '$DBHandler','$DBName','$DBUser','$DBPassword'\n";
	my $self = 	{ Name=>$DBName, job=>$job } ;
	bless $self, $class ;
	my $ccDB = $self->{ccDB} = ezApp::DB->new($job, $DBName, $DBUser, $DBHandler, $DBPassword) ;
	$self->{isopen} =  $ccDB->{isopen} && $self->FetchPrimaryHashes ;		
$self}
#-----------------------------------------------------------------------------------------------------------
sub FetchPrimaryHashes 			# Read in hashes for record sizes, currencies, & origins.
{ 
	my $self = shift ;
	my $job = $self->{job} ;
	my $ccDB = $self->{ccDB} ;
	
	# Setup currency owners hash.
	
		my $Rows = 
			$ccDB->Rows('Currencies', -Columns=>['CurrencyId','CreatorAcctId'])
				|| return $job->Warn ("Cannot read Currencies table") ;
				
		$self->{CurrencyOwnerIds} = { map { $_->[0] => $_->[1] } @$Rows } ;
			
	# Setup acct passwords hash.
	
		$Rows = 
			$ccDB->Rows('Origins', -Columns=>['AcctId', 'AcctPassword'])
				|| return $job->Warn ("Cannot read Origins table") ;
		$self->{UserPasswords} = { map { $_->[0] => $_->[1] } @$Rows } ;
			
1}
#-----------------------------------------------------------------------------------------------------------
sub IdsOk 							# Convert list of Id's to lc and verify that 
												# they are alphanum with no spaces.
{ 
	my ($self, @IdRefs ) = @_ ;
	my $job = $self->{job} ;
	foreach my $IdRef (@IdRefs) {
		$$IdRef = lc $$IdRef ;
		$$IdRef =~ /^[a-z0-9]{3,20}$/ 		
			|| return $job->Warn 
				("'$$IdRef' invalid: Between 3 and 20 letters and digits allowed, with no spaces") ;
	} ;
1}
#-----------------------------------------------------------------------------------------------------------
sub AddCurrency					# Add new currency record.  Set all columns other than CurrencyId to defaults.
{
	my ($self, $CurrencyId ) = @_ ;	

	my ( $job, $ccDB, $CurrencyOwnerIds, $UserPasswords ) = @$self{ 
		qw( job ccDB CurrencyOwnerIds UserPasswords ) } ;
		
	# Verify validity of CurrencyId.
		$self->IdsOk( \$CurrencyId ) || return '' ;
		$CurrencyOwnerIds->{$CurrencyId}
			&& return $job->Warn("'$CurrencyId' is already an established currency")  ;
			
	# Insert new Currencies record.
		$ccDB->DoSQL("insert into Currencies (CurrencyId) values ('$CurrencyId')") 
			|| return '' ;
	
$self->FetchPrimaryHashes }
#-----------------------------------------------------------------------------------------------------------
sub AddOrigin 					# Add new user, setting cols other than AcctId and AcctPassword to defaults.
{ 
	my ($self, $AcctId, $AcctPassword ) = @_ ;
	
	my ( $job, $ccDB, $UserPasswords, $CurrencyOwnerIds ) = @$self{ 
		qw( job ccDB UserPasswords CurrencyOwnerIds ) } ;
		
	# Verify validity of arguments.
		( $AcctId && $AcctPassword ) 
			|| return $job->Warn('$AcctId and $AcctPassword required by ccKernel->AddOrigin') ;
		$self->IdsOk( \$AcctId, \$AcctPassword ) 
			|| return '' ;
		$UserPasswords->{$AcctId} 
			&& return $job->Warn("'$AcctId' is already an established account") ;
	
	# Insert new Origins record.
		$ccDB->DoSQL
			("insert into Origins ( AcctId, AcctPassword ) values ( '$AcctId', '$AcctPassword' )")
				|| return '' ;
	
$self->FetchPrimaryHashes }
#-----------------------------------------------------------------------------------------------------------
sub CreateLedger		# Create ledger record for specified account and currency.
										# Set to default any fields other than AcctId and CurrencyId.
{
	my ( $self, $AcctId, $CurrencyId ) = @_ ;
	
	my ( $ccDB, $UserPasswords, $CurrencyOwnerIds, $job ) = 
		@$self{ qw( ccDB UserPasswords CurrencyOwnerIds job ) } ;
		
	( $UserPasswords->{$AcctId} && $CurrencyOwnerIds->{$CurrencyId} )
		|| return $job->Warn('ccKernel->CreateLedger given invalid arguments') ;
	my $Rows = $ccDB->Rows
		( 'Ledgers', -Columns=>['AcctId'], -Where=>"AcctId='$AcctId' and CurrencyId='$CurrencyId'" )
			|| return $job->Warn('Could not read Ledgers') ;
	@$Rows && return $job->Warn("$AcctId already registered for $CurrencyId") ;
	$ccDB->DoSQL
		( "insert into Ledgers (AcctId, CurrencyId) values ('$AcctId', '$CurrencyId')" )
			|| return $job->Warn("Could not create Ledger for $AcctId") ;
1}
#-----------------------------------------------------------------------------------------------------------
sub Transfer  			# Record transaction in Acks & Ledgers.
{
	my ($self, $SourceAcctId, $DestAcctId, 
	    $CurrencyId, $Amount, $Item) = @_ ;
	    
	my ( $ccDB, $UserPasswords, $CurrencyOwnerIds, $job ) = @$self{ 
		qw( ccDB UserPasswords CurrencyOwnerIds job ) } ;
	
	# Verify validity of arguments.
		$UserPasswords->{$SourceAcctId} 
			|| return $job->Warn("'$SourceAcctId' not a valid account") ;
		$UserPasswords->{$DestAcctId} 
			|| return $job->Warn("'$DestAcctId' not a valid account") ;
		$SourceAcctId ne $DestAcctId 
			|| return $job->Warn("Source and destination accounts are the same") ;
		$CurrencyOwnerIds->{$CurrencyId} 
			|| return $job->Warn("'$CurrencyId' is not a valid currency") ;
		$Amount =~ /^-?[0-9]+$/
			|| return $job->Warn("Amount must be an integer") ;
			
	# Verify that ledgers exist for both users in this currency.
		my $Rows = $ccDB->Rows
			(
				'Ledgers', 
				-Columns=>['AcctId'], 
				-Where=>"CurrencyId='$CurrencyId' and ( AcctId='$SourceAcctId' or AcctId='$DestAcctId' )"
			)
				|| return $job->Warn('Cannot read Ledgers') ;
		2 == @$Rows
			|| return $job->Warn("Trade invalid: both accounts must be registered for $CurrencyId") ;
		
	# Update ledger records and get new balances.
		my $TradeSign = $Amount > 0 ? '+' : '-' ;
		my $UpdateSourceLedger = "update Ledgers set Balance=Balance-$Amount, TradesTotal=TradesTotal+$Amount, "
			. "TradesCount = TradesCount $TradeSign 1, LastTradeDate=now() "
			. "where AcctId='$SourceAcctId' and CurrencyId='$CurrencyId'" ;
		my $UpdateDestLedger = "update Ledgers set Balance=Balance+$Amount, TradesTotal=TradesTotal+$Amount " .
			"where AcctId='$DestAcctId' and CurrencyId='$CurrencyId'" ;
		my ( $SourceBalance, $DestBalance ) ;
		( $ccDB->DoSQL($UpdateSourceLedger)
			&& $ccDB->DoSQL($UpdateDestLedger)
			&& 
			( 
				$SourceBalance = $ccDB->Rows
					('Ledgers', -Columns=>['Balance'], -Where=>"AcctId='$SourceAcctId' and CurrencyId='$CurrencyId'") 
			)
			&&
			( 
				$DestBalance = $ccDB->Rows
					('Ledgers', -Columns=>['Balance'], -Where=>"AcctId='$DestAcctId' and CurrencyId='$CurrencyId'") 
			)
		)
			|| return $job->Warn('Ledgers could not be updated and are now inconsistent') ;
		( $SourceBalance, $DestBalance ) = map { shift @{ shift @$_ } } 
			( $SourceBalance, $DestBalance ) ;
		
	# Write the Acks record.
		my $Cols = 
			'AckDateTime, SourceAcctId, DestAcctId, AckCurrencyId, AckAmount, AckItem, SourceBalance, DestBalance';
		my $Values = "now(), " . $ccDB->QuotedSQLlist
			( [ $SourceAcctId, $DestAcctId, $CurrencyId, $Amount, $Item, $SourceBalance, $DestBalance ] ) ;	
		$ccDB->DoSQL
			("insert into Acks ($Cols) values ($Values)") 
				|| return $job->Warn('Trade not recorded: ledgers inconsistent') ;
1}
#-----------------------------------------------------------------------------------------------------------
1 ;
#-----------------------------------------------------------------------------------------------------------



