#!/usr/bin/perl -w
use diagnostics;
use strict;

use HTML::Entities();
use FileHandle();

package ezApp::Compiler;
	# subclass of HTML::Parser.
		use HTML::Parser; 
		use vars qw(@ISA);	
		@ISA = qw(HTML::Parser);

=comment
-------------------------------------------------------------------

		Copyright (c) 2001 Richard K. Moore.
				richard@cyberjournal.org

 This script is free software; you can redistribute it
 and/or modify it under the same terms as Perl itself.

-------------------------------------------------------------------

Function
^^^^^^^^

	Our functional method is &CompileFile($File).
	
	We compile the source file into a tree of blocks, each block
	being the content at a single level of nesting with respect
	to the ezApp control tags: <if> <while> </if>, etc.  Each
	block has a key, and the block is stored in $Blocks->{key}.
	The topmost block has key value 1.
	
	Our output is an encoded representation of the $Blocks hash,
	given by 
		join "\n".chr(3)."\n", 'key1 =>', block1,'key2=>', block2....
		
	"\n" enables the compiled file to be viewed conveniently in
	a text editor, without adding appreciably to the length of
	the file.

Implementation
^^^^^^^^^^^^^^

	Using HTML::Parser, we scan through the items of our source
	file. An 'item' is a comment, declaration, text, start tag,
	or end tag. Comments and declarations are processed as if
	they were text items.
	
	Each item is encoded into a string by joining its tokens
	with chr(1). Blocks are created by joining items together
	with chr(2)."\n".  
	
	The tokens of an item occur in this order:
				
			1) source line number at start of item, or '' if the line number was already emitted
			2) 
				'<'  	=> start tag, 
				'/' 	=> end tag, 
				'"' 	=> text.
			3) tag name, if not text, else text value
			4) for block items: key of nested block
			5) for all start-tag items: (attrname1, attrvalue1, attrname2,...) 
	
-------------------------------------------------------------------
=cut
		

	# Hashes to encode tagnames as integers, if that becomes advisable.
	# Meanwhile, they serve as hashes to test what kind of tag we've got.
	
		my @ezBlockTags = qw
			( if elsif else while repeat foreach subscreen screen );
			
		my $tagKeyVal = 5;
		my $ezBlockTagCodes = 
			{ map {$_ => $tagKeyVal++} @ezBlockTags };
			
	
		my @ezNonBlockTags = qw
		( 
			editvar persistent nonpersistent hiddenfields 
			ezbutton gotoscreen
			dbopen dbclose dbgetrecords dbupdate dbdo dblastinsertid
		);
	
		my $ezNonBlockTagCodes = 
			{ map {$_ => $tagKeyVal++} @ezNonBlockTags };
			
	
		my @HtmlTags = qw
		(
			a abbr acronym address applet area b basefont base bdo bgsound
			big blink blockquote body br button caption center cite code col
			colgroup dd del dir div dl dt em embed fieldset font form
			frame frameset h1 h2 h3 h4 h5 h6 head hr html i iframe img
			input ins isindex kbd keygen label layer legend li link map
			marquee meta menu multicol nobr nembed noframes nolayer
			noscript object ol optgroup option p param pre q s samp
			script select server small spacer span strike strong
			sub sup table tbody td textarea tfoot th thead title tr
			tt u ul var wbr
		);
	
		my $HtmlTagCodes = 
			{ map {$_ => $tagKeyVal++} @HtmlTags };
			
		my $TagCodes = {%$ezBlockTagCodes, %$ezNonBlockTagCodes, %$HtmlTagCodes};

#-------------------------------------------------------------------

sub new
{
	my $class = shift;
	my $self = HTML::Parser->new; 
	bless $self, $class;	
$self}


sub dieClean
{
	my ($self, $message) = @_;
	$message .= $self->{LineNumber} 
		? " line $self->{LineNumber} of file '$self->{File}'" 
		: '';
	die "$message.\n";
}


	my $MatchingEndTags =
		{ 'elsif'=>'if', 'else'=>'if', map { $_ => $_ } 'if', 'while','repeat','foreach','subscreen', 'screen' };
		
sub CompileFile
{
	my $self											= shift;
	my $File = $self->{File}			= shift;
	$self->{LineNumber}	 					= 0;
	$self->{NestingStack}					= [];
	$self->{NextKey}							= 1;
	$self->{Key}									= $self->{NextKey}++;
	$self->{Blocks}								= {};
	$self->{LastShownLineNumber}	= '';
	$self->{InScreen}							= '';
	$self->{ScreenSeen}						= '';
	$self->{Subscreens}						= {};
	
	my $FileContents = $self->ExpandFile($File);
	
	my $OutFile = $File =~ /^(.+?).scr.txt$/s
		? "ezscrs/$1.ezscr"
		: $self->dieClean("Illegal file name: '$File'");
		
	(open OUT, ">$OutFile")
		|| $self->dieClean("Could not open '$OutFile' for output: $!");
		
	$self->{LineNumber}++;
	$self->parse($FileContents);
	$self->eof;
	
	my $DanglingFrame = pop @{ $self->{NestingStack} };
	if ($DanglingFrame)
	{
		my ($DanglingNestingTag, $DanglingKey) = @$DanglingFrame;
		my $DanglingBlock = $self->{Blocks}->{$DanglingKey};
		my @Items = split chr(2)."\n", $DanglingBlock;
		my $FirstItem = shift @Items;
		my @Tokens = split chr(1), $FirstItem;
		$self->{LineNumber} = shift @Tokens;
		$self->{LineNumber} = $1
			if $self->{LineNumber} =~ /^.+?\t(.+?)$/s;
		$self->dieClean("Unterminated <$DanglingNestingTag> at");
	};
	
	my $Blocks = $self->{Blocks};
	print OUT  join "\n".chr(3)."\n", 
		map { ( "$_", $Blocks->{$_} ) } keys %$Blocks;
	close OUT;
}


sub CompileItem
{
	my ($self, $rawtext, $Kind, $TagOrText, @tokens) = @_;
	
	my $LineNumber = $self->{LineNumber};
	my $LineNumberField =  $LineNumber eq $self->{LastShownLineNumber}
		? ''
		: $LineNumber;
	$self->{LastShownLineNumber} = $LineNumber;
	
	$TagOrText = $self->CompileText($TagOrText)
		if $Kind eq '"';
	
	
			# Tag-compression code. Is it worth it? Also we can check here 		**********
			# for valid tags
				#	my $TagField = $Kind eq '"'
				#		? $Tag
				#		: $TagCodes->{$Tag};
	
	my $Key = $self->{Key};
	my $BlockRef = \$self->{Blocks}->{$Key};
	$$BlockRef = join chr(2)."\n", 
		( $$BlockRef ? $$BlockRef : () ), join chr(1), 
			$LineNumberField, $Kind, $TagOrText, @tokens;
	$self->AddLineCount($rawtext);
}


sub	 AddLineCount
{
	my ($self, $rawtext) = @_;
	my @LineEnds = $rawtext =~ m"\n"sg;
	$self->{LineNumber} += @LineEnds;
}


sub text
{
		my ($self, $text) = @_;
		
		$text !~ /^\</
			|| return $self->dieClean("HTML syntax error at");
		
		$self->TextEtc($text);
}


sub TextEtc
{
		my ($self, $TextEtc) = @_;
		$self->CompileItem($TextEtc, '"', $TextEtc);
}


sub comment				# Pass through comments as text.
{
	my ($self, $comment) = @_;
	
	$self->TextEtc("<!--$comment-->");
	
	if ( $comment =~ /^#use (.+?)$/s )
	{
		$self->{File} = $1;
		push @{ $self->{LineNumberStack} }, $self->{LineNumber};
		$self->{LineNumber} = 1;
	};
	
	if ( $comment =~ /^#resume (.+?)$/s )
	{
		$self->{File} = $1;
		$self->{LineNumber} = pop @{ $self->{LineNumberStack} }
			|| $self->dieClean("empty LineNumberStack");
	};
}


sub declaration 	# Same with declarations.
{
	my ($self, $declaration) = @_;
	$self->TextEtc("<!$declaration>");
}


sub end
{ 
	my ($self, $tag) = @_;
	
	$self->{InScreen} = ''
		if $tag eq 'screen';
	
	$ezBlockTagCodes->{$tag}
		|| return $self->CompileItem("</$tag>", '/', $tag);
	
	( my $NestingFrame = pop @{ $self->{NestingStack} } )
		or die "Compiler: NestsingStack empty";
		
	my ( $NestingTag, $Key, $File, $LineNumber ) = @$NestingFrame;
	
	unless ( $NestingTag  && (defined $Key) && $tag eq $MatchingEndTags->{$NestingTag} )
	{
		@$self{'File','LineNumber'} = ($File, $LineNumber);
		$self->dieClean("No terminator found for <$NestingTag> at");
	};

	$self->{Key} = $Key;
$NestingTag}


sub start
{
	my $self = shift;
	my ($tag, $attr, $attrseq, $tagtext) 	= @_;
	$attrseq 															||= [];
	my @tokens 														= map { $_, $attr->{$_} } @$attrseq;
	
	$TagCodes->{$tag} || $self->{Subscreens}->{$tag}
		|| $self->dieClean("Unrecognized tag <$tag> at");
	
	if ($tag eq 'screen')
	{
		! $self->{ScreenSeen}++
			|| $self->dieClean("Duplicate <screen> tag at");
		$self->{InScreen}++;
	};
	
	$self->{InScreen}
		|| $self->dieClean("<$tag> is not enclosed withn a <screen> block at");
		
	if ( $tag =~ /^(else|elsif)$/ )
	{
	
		my $NestingTag = $self->end('if');
		$NestingTag =~/^(if|elsif)$/
			|| $self->dieClean("Following <$NestingTag>, <$tag> is improper at");
		
	};
	
	if ( $self->{Subscreens}->{$tag} )
	{
		unshift @tokens, $self->{Subscreens}->{$tag};
	}
	elsif ( $ezBlockTagCodes->{$tag} )
	{
		if ( $tag eq 'subscreen' )
		{
			$self->dieClean("<subscreen> requires 'name' as initial attribute at")
				unless $tokens[0] eq 'name';
			my $SubscreenName = $tokens[1];
			$self->dieClean("Attempt to redefine '$SubscreenName' at")
				if $TagCodes->{$SubscreenName} || $self->{Subscreens}->{$SubscreenName};
			$self->{Subscreens}->{$SubscreenName} = $self->{NextKey};
		};
		
		unshift @tokens, $self->{NextKey};
	};
		

	# Compile our possibly expanded item.
	
		$self->CompileItem($tagtext, '<', $tag, @tokens);	
		return 
			unless $ezBlockTagCodes->{$tag};
		
	
	# Enter new nesting level to compile sub block.
	
		push @{ $self->{NestingStack} }, [ $tag, $self->{Key}, $self->{File}, $self->{LineNumber} ];
		$self->{Key} = $self->{NextKey}++;
	
}


sub CompileText		# Process {{...}}, '{...}',  "{...}".
{
	my ($self, $Text) = @_;
	my @Snips = ();
	while ( $Text =~ /^(.*?)([{]|[']|["])[{](.+)$/s )
	{
		push @Snips, $1;
		my $BraceOpenBracket = $2;
		my $BraceCloseBracket = $2 eq '{'
			? '}'
			: $2;
		$Text = $3;
		( $Text =~ /^(.*?)[}][$BraceCloseBracket](.*?)$/s )
			|| $self->dieClean
				("Unmatched braces - open = '$BraceOpenBracket', close = '$BraceCloseBracket', Text ='$Text' - in text at");
		push @Snips, "$BraceOpenBracket\{$1}$BraceCloseBracket";
		$Text = $2;
	};
	
	$Text = join chr(4), @Snips, $Text;
$Text}


sub ExpandFile	# Incorporate any includes.
{
	my ($self, $FileName) = @_;
	my $html							='';
	
	my ($FileContents, $Error);
	$self->ReadEntireFile("screens/$FileName", \$FileContents, \$Error);
	! $Error
		|| $self->dieClean($Error);
		
	while ( $FileContents =~ /(.+?)(<!--#use (.+?)-->)?/sg )
	{
		$html .= $1;
		if ($3)
		{
			my $InnerFileName = $3;
			$html .= " $2 " . $self->ExpandFile($InnerFileName) . " <!--#resume $FileName--> ";
		};
	};

$html}


sub ReadEntireFile			# Read file into ContentsRef and return 1.
												# If problems, return '' and set ErrorMessage.
{	
	my ( $self, $FileName, $ContentsRef, $ErrorMessageRef ) = @_  ;
	if ( ! open IN, "<$FileName" ) {
		$$ErrorMessageRef = "Cannot open $FileName for input: $!" ;
		return '' ;
	} ;
	my $OldSep = $/ ; undef $/ ; 				## Mask $/ to read in one slurp
		$$ContentsRef = <IN> ;
	undef $/ ; $/ = $OldSep ;						## Restore record terminator and correct BBEdit sytax coloring.
	close IN ;
1 } 


1;	
